import abc
import os
import random
import itertools

from .sh import run_cmd


class CaseSet:

    def __init__(self):
        self.items = []

    def __iter__(self):
        self._iter = iter(self.items)

    def __next__(self):
        return next(self._iter)


class FileCaseSet(CaseSet):

    def __init__(self):
        self.items = [
            NonexistentFileCase(),
            EmptyFileCase(),
        ]


class TmpDir:

    def __init__(self, path='tmp'):
        self.path = path
        run_cmd(['mkdir', '-p', self.path])

    def destroy(self):
        run_cmd(['rm', '-r', self.path])


class FileCase(abc.ABC):

    name = '_unknown_'
    writable = True

    def __init__(self, root, content=None):
        self.root = root
        self._props = {}
        self._props.update(self.build_props())
        self._base_content = content or ''

    def build_props(self):
        return {}

    def __write(self):
        data = ''
        for ln in self.build_lines():
            data += (ln + '\n')
        data += self.build_text()
        with open(self.path, 'w') as fh:
            fh.write(data)

    def create(self):
        if not self.writable:
            return
        self.__write()

    def destroy(self):
        if not self.writable:
            return
        os.remove(self.path)

    def exists(self):
        return os.path.exists(self.path)

    @property
    def path(self):
        return self.root + '/' + self.name

    def build_lines(self):
        if type(self._base_content) is list:
            return self._base_content
        else:
            return []

    def build_text(self):
        if type(self._base_content) is str:
            return self._base_content
        else:
            return ''


class NonexistentFileCase(FileCase):

    name = 'nonexistent'
    writable = False

    def build_props(self):
        return {'id': random.randint(10000000, 99999999)}

    @property
    def path(self):
        rid = self._props["id"]
        return f'/tmp/nonexistent_file_{rid}'


class EmptyFileCase(FileCase):

    name = 'empty'


class Type(abc.ABC):

    def __init__(self, idx=0):
        self.idx = idx
        self.val = self[idx]

    @staticmethod
    @abc.abstractmethod
    def all_values():
        return []

    @classmethod
    def idx_range(cls, fill=False):
        idxs = [v[0] for v in cls.all_values()]
        if fill:
            return idxs
        return [None] + idxs

    def __getitem__(self, idx):
        for sidx, sval in self.all_values():
            if sidx == idx:
                return sval
        raise IndexError(f'index out of range: {idx!r}')

    def next(self):
        return self.__class__(self.idx + 1)

    def prev(self):
        return self.__class__(self.idx - 1)

    def window(self, size=1, offset=0):
        win = []
        win_idx = 0
        while size - win_idx:
            win.append(self[win_idx+offset])
            win_idx = win_idx + 1
        return win


class GrowingCharType(Type):

    @staticmethod
    def all_values():
        return [
            (-3, 'j'),
            (-2, 'k'),
            (-1, 'l'),
            (0, 'm'),
            (1, 'n'),
            (2, 'o'),
            (3, 'p'),
        ]


class GrowingWordType(Type):

    @staticmethod
    def all_values():
        return [
            (-3, 'jack'),
            (-2, 'king'),
            (-1, 'luna'),
            (0, 'mark'),
            (1, 'neon'),
            (2, 'orbs'),
            (3, 'pink'),
        ]

    @classmethod
    def idx_range(cls):
        return [None, 0, 1, -1, 2, -2, 3, -3]


class ArgOrValueType(Type):

    @staticmethod
    def all_values():
        return list(enumerate([
            '--foo',
            '--foobar',
            '--bar',
            'bar',
            'foo',
            '-f',
            '',
            '-',
            '--',
            'foobar',
            '-fb',
        ]))


class WordType(Type):

    @staticmethod
    def all_values():
        return list(enumerate([
            'foo',
            'bar',
            'baz',
            'qux',
            'quux',
            'foobar',
        ]))


class ValueCase:

    def __init__(self, val):
        self.val = val

    # O='foo'
    # R='food'
    # size=False
    # atOa=True
    # atRa=False
    # atOz=False
    # atRz=False

    # O='ood'
    # R='food'
    # size=False
    # atOa=False
    # atRa=False
    # atOz=False
    # atRz=False


class ScalarCaseSet(CaseSet):

    def __init__(self, tpe):
        def gen():
            while True:
                yield tpe.next().value
        self.tpe = tpe
        self.items = []
        self.matching = None
        for idx, val in tpe.all_values():
            self.items.append(val)
            if idx == 0:
                self.matching = val


_IDX_RANGE = [None, 0, 1, -1, 2, -2, 3, -3]

_DEFAULT_REP = 5


def gen_idxs(idx_range, maxsize=None, maxitems=None):

    def prune(L):
        return tuple([x for x in L if x is not None])

    def product():
        seen = set()
        rep = maxsize if maxsize is not None else _DEFAULT_REP
        for dirty in itertools.product(idx_range, repeat=rep):
            clean = prune(tuple(dirty))
            if clean in seen:
                continue
            yield clean
            seen.add(clean)

    def limit():
        if maxitems is None:
            return product()
        return itertools.islice(product(), maxitems)

    return limit()


class ArrayCaseSet(CaseSet):

    def _gen_items(self):
        igen = gen_idxs(
            idx_range=self.tpe.idx_range(),
            maxsize=self.maxsize,
            maxitems=self.maxitems,
        )
        for idxset in igen:
            item = [self.tpe[i] for i in idxset]
            yield item

    def __init__(self, tpe, maxsize=None, maxitems=None):
        self.tpe = tpe
        self.maxsize = maxsize
        self.maxitems = maxitems
        self.items = list(self._gen_items())
