import json
import difflib


class _NL:

    @staticmethod
    def split(text):
        tmp = text
        if text[-1] == '\n':
            tmp = text[:-1]
        return tmp.split('\n')

    @staticmethod
    def join(lines):
        return '\n'.join(lines) + '\n'

    @staticmethod
    def maybe_join(lines):
        if type(lines) is list:
            return '\n'.join(lines) + '\n'
        elif type(lines) is str:
            return lines
        else:
            return str(lines)


class _ValueHolder:

    def __init__(self, value):
        self.value = value

    def __eq__(self, othr):
        return self.value == othr.value


class Oracle(_ValueHolder):
    """
    Prediction of how the SUT should behave.

    This class does not contain any methods to *evaluate* the result, it's
    simply a bearer of the information.

    *value* holds values expected from the SUT and must be investigated by
    specific assert functions.

    Optionally, *exc* can hold exception class if SUT should throw this
    exception.

    Any extra keyword arguments *kws* can hold other non-functional behavior
    parameters and can be accessed directly using dot notation or getattr().
    """

    def __init__(self, value, exc=None, **kws):
        self.value = value
        self.exc = exc
        self._flags = kws
        for key, value in kws.items():
            setattr(self, key, value)

    def flag(self, flag):
        if flag in self._flags:
            return self._flags[flag]
        raise KeyError(f"no such flag: {flag!r}")

    def __str__(self):
        parts = []
        for key in self._flags:
            parts.append(key + '=' + str(self._flags[key])[0])
        attrs = ','.join(parts)
        cn = self.__class__.__name__
        return f'{cn}(value={self.value},{attrs})'


class Result(_ValueHolder):
    """
    Result data holder.

    This class does not contain any methods to *evaluate* the result, it's
    simply a bearer of the information.

    *value* holds any return values from test function.

    This should later be investigated by assert functions.
    """
    pass


class ShellOracle:

    def __init__(self, out=None, err=None, es=None):
        self.out = _NL.maybe_join(out) if out else ''
        self.err = _NL.maybe_join(err) if err else ''
        self.es = es if es else 0


class ShellResult:

    @staticmethod
    def _normalize_tb(err):
        if not err.startswith('Traceback (most recent call last):'):
            return err
        out = []
        lines = _NL.split(err)
        out.append(lines.pop(0))
        while lines:
            if lines[0].startswith('  File'):
                lines.pop(0)
                continue
            elif lines[0].startswith('    '):
                lines.pop(0)
                continue
            else:
                break
        out.extend(lines)
        return _NL.join(out)

    @classmethod
    def normalize_shresult(cls, shresult):
        es = shresult.es
        out = shresult.out
        err = cls._normalize_tb(shresult.err)
        return cls(out=out, err=err, es=es)

    def __init__(self, out=None, err=None, es=None):
        self.out = _NL.maybe_join(out) if out else ''
        self.err = _NL.maybe_join(err) if err else ''
        self.es = es if es else 0
