#!/usr/bin/python3


import subprocess
import neaty.log


def _noop(*args, **kwargs):
    pass


def _join_lines(lines):
    return ''.join([ln + '\n' for ln in lines])


class _ShellResult:

    def __init__(self, out, err, es, action):
        self.action = action
        self.out = out
        self.err = err
        self.es = es

    def is_failure(self):
        return self.es > 0

    def is_success(self):
        return self.es == 0

    def _fmt_payload(self):
        return '(something)'

    def _log_failing(self):
        neaty.log.warn(self._fmt_payload())
        for line in self.out.splitlines():
            neaty.log.warn("  out:|" + line)
        for line in self.err.splitlines():
            neaty.log.warn("  err:|" + line)

    def _log_successful(self):
        neaty.log.think(self._fmt_payload())
        for line in self.err.splitlines():
            neaty.log.warn(line)

    def log(self):
        fn = {0: self._log_successful}.get(self.es, self._log_failing)
        fn()


class CommandResult(_ShellResult):

    def _fmt_payload(self):
        verb = "succeeded"
        if self.is_failure():
            verb = "exited with: %d" % self.es
        return ("%s: command finished: %s (unquoted here); %s"
                % (self.action.name, verb, " ".join(self.action.args)))


class ScriptResult(_ShellResult):

    def _fmt_payload(self):
        verb = "succeeded"
        if self.is_failure():
            verb = "exited with: %d" % self.es
        return "%s: script finished: %s" % (self.action.name, verb)


class _ShellAction:

    def run(self):
        if self.dry:
            return self.result_cls(
                action=self,
                out='',
                err='',
                es=0,
            )
        p = subprocess.Popen(
                self.args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
        )
        out, err = p.communicate()
        return self.result_cls(
            action=self,
            out=out.decode(),
            err=err.decode(),
            es=p.returncode,
        )


def stdout_of(cmd):
    result = ShellCommand(cmd, name='stdout_of()').run()
    if not result.is_success():
        raise RuntimeError("command failed")
    return result.out


def run_cmd(cmd, name):
    result = ShellCommand(cmd, name).run()
    result.log()
    return result.is_success()


def run_code(code, name):
    result = ShellScript(code, name).run()
    result.log()
    return result.is_success()


def dry_run_code(code, name):
    neaty.log.warn("-- begin code %s --" % name)
    for line in code.split('\n'):
        neaty.log.warn('|' + line.rsplit('\n'))
    neaty.log.warn("-- end code %s --" % name)


def quote(text):
    return stdout_of([
        'bash', '-c', 'printf %q "$1"',  '--', text
    ])


def content_equals(path_a, path_b):
    with open(path_a) as f:
        ca = f.read()
    with open(path_b) as f:
        cb = f.read()
    return ca == cb


class ShellCommand(_ShellAction):

    def __init__(self, cmd, name, dry=False, verbose=False, debug=False):
        self.args = cmd
        self.name = name
        self.dry = dry
        self.verbose = verbose
        self.debug = debug
        self.result_cls = CommandResult


class ShellScript(_ShellAction):

    def __init__(self, code, name, dry=False, verbose=False, debug=False):
        self.args = ['sh', '-c', code]
        self.name = name
        self.dry = dry
        self.verbose = verbose
        self.debug = debug
        self.result_cls = ScriptResult
